import glob
import sys
from typing import List

import click
import uvicorn
from rdflib import Dataset

from rdflib_endpoint import SparqlEndpoint


@click.group()
def cli() -> None:
    """Quickly serve RDF files as SPARQL endpoint with RDFLib Endpoint"""


@cli.command(help="Serve a local RDF file as a SPARQL endpoint by default on http://0.0.0.0:8000/sparql")
@click.argument("files", nargs=-1)
@click.option("--host", default="localhost", help="Host of the SPARQL endpoint")
@click.option("--port", default=8000, help="Port of the SPARQL endpoint")
@click.option("--store", default="default", help="Store used by RDFLib: default or Oxigraph")
@click.option("--enable-update", is_flag=True, help="Enable SPARQL updates")
def serve(files: List[str], host: str, port: int, store: str, enable_update: bool) -> None:
    run_serve(files, host, port, store, enable_update)


def run_serve(files: List[str], host: str, port: int, store: str = "default", enable_update: bool = False) -> None:
    if store == "oxigraph":
        store = store.capitalize()
    g = Dataset(store=store, default_union=True)
    for glob_file in files:
        file_list = glob.glob(glob_file)
        for file in file_list:
            g.parse(file)
            click.echo(
                click.style("INFO", fg="green")
                + ":     📥️ Loaded triples from "
                + click.style(str(file), bold=True)
                + ", for a total of "
                + click.style(str(len(g)), bold=True)
            )

    app = SparqlEndpoint(
        graph=g,
        enable_update=enable_update,
        example_query="""PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
SELECT * WHERE {
    ?s ?p ?o .
} LIMIT 100""",
    )
    uvicorn.run(app, host=host, port=port)


@cli.command(help="Merge and convert local RDF files to another format easily")
@click.argument("files", nargs=-1)
@click.option("--output", default="localhost", help="Host of the SPARQL endpoint")
@click.option("--store", default="default", help="Store used by RDFLib: default or Oxigraph")
def convert(files: List[str], output: str, store: str) -> None:
    run_convert(files, output, store)


def run_convert(files: List[str], output: str, store: str = "default") -> None:
    if store == "oxigraph":
        store = store.capitalize()
    g = Dataset(store=store, default_union=True)
    for glob_file in files:
        file_list = glob.glob(glob_file)
        for file in file_list:
            g.parse(file)
            click.echo(
                click.style("INFO", fg="green")
                + ":     📥️ Loaded triples from "
                + click.style(str(file), bold=True)
                + ", for a total of "
                + click.style(str(len(g)), bold=True)
            )

    out_format = "ttl"
    if output.endswith(".nt"):
        out_format = "nt"
    elif output.endswith(".xml") or output.endswith(".rdf"):
        out_format = "xml"
    elif output.endswith(".json") or output.endswith(".jsonld"):
        out_format = "json-ld"
    elif output.endswith(".trig"):
        out_format = "trig"

    g.serialize(output, format=out_format)


if __name__ == "__main__":
    sys.exit(cli())



import time
from multiprocessing import Process

import httpx
import pytest
import uvicorn
from example.main import custom_concat
from rdflib import Graph

# from testcontainers.core.container import DockerContainer
# from testcontainers.core.waiting_utils import wait_for_logs
from rdflib_endpoint import SparqlEndpoint

# https://github.com/biopragmatics/curies/blob/main/tests/test_federated_sparql.py

graph = Graph()


def _get_app():
    return SparqlEndpoint(
        graph=graph,
        functions={
            "https://w3id.org/sparql-functions/custom_concat": custom_concat,
        },
        enable_update=True,
    )


@pytest.fixture(scope="module")
def service_url():
    host = "localhost"
    port = 8000
    service_process = Process(
        target=uvicorn.run,
        args=(_get_app,),
        kwargs={"host": host, "port": port, "log_level": "info"},
        daemon=True,
    )
    service_process.start()
    time.sleep(2)
    endpoint_url = f"http://{host}:{port}"
    yield endpoint_url
    service_process.kill()
    service_process.join()


# Stop and delete all testcontainers: docker stop $(docker ps -a -q) && docker rm $(docker ps -a -q)
# NOTE: in case issue in rootless docker: https://github.com/testcontainers/testcontainers-python/issues/537
# TESTCONTAINERS_DOCKER_SOCKET_OVERRIDE=/run/user/$(id -u)/docker.sock uv run pytest tests/test_federation.py -s

# import os

# env = os.environ.copy()
# env["GRAPHDB_USERNAME"] = "admin"
# env["GRAPHDB_PASSWORD"] = "root"

# @pytest.fixture(scope="module")
# def graphdb():
#     """Start GraphDB container as a fixture."""
#     container = DockerContainer("ontotext/graphdb:10.8.4")
#     container.with_exposed_ports(7200).with_bind_ports(7200, 7200)
#     container.with_env("JAVA_OPTS", "-Xms1g -Xmx4g")
#     container.start()
#     delay = wait_for_logs(container, "Started GraphDB")
#     base_url = f"http://{container.get_container_host_ip()}:{container.get_exposed_port(7200)}"

#     print(f"GraphDB started in {delay:.0f}s at {base_url}")
#     # print(container.get_logs())
#     yield base_url


# def test_graphdb_custom_concat(service_url, graphdb):
#     print(concat_select.format(rdflib_endpoint_url=service_url))
#     response = httpx.get(graphdb, params={"query": concat_select.format(rdflib_endpoint_url=service_url)}, headers={"accept": "application/json"})
#     print(response.text)
#     assert response.status_code == 200
#     assert response.json()["results"]["bindings"][0]["concat"]["value"] == "Firstlast"


# @pytest.fixture(scope="module")
# def blazegraph():
#     """Start blazegraph container as a fixture."""
#     container = DockerContainer("lyrasis/blazegraph:2.1.4")
#     container.with_exposed_ports(8080).with_bind_ports(8080, 8080)
#     # container.with_env("JAVA_OPTS", "-Xms1g -Xmx4g")
#     container.start()
#     delay = wait_for_logs(container, "Started @")
#     base_url = f"http://{container.get_container_host_ip()}:{container.get_exposed_port(8080)}/bigdata/namespace/kb/sparql"

#     print(f"Blazegraph started in {delay:.0f}s at {base_url}")
#     # print(container.get_logs())
#     yield base_url


# def test_blazegraph_custom_concat(service_url, blazegraph):
#     print(concat_select.format(rdflib_endpoint_url=service_url))
#     response = httpx.get(blazegraph, params={"query": concat_select.format(rdflib_endpoint_url=service_url)}, headers={"accept": "application/json"})
#     print(response.text)
#     assert response.status_code == 200
#     assert response.json()["results"]["bindings"][0]["concat"]["value"] == "Firstlast"


def test_direct_custom_concat(service_url):
    direct_concat_select = """PREFIX myfunctions: <https://w3id.org/sparql-functions/>
SELECT ?concat WHERE {
        BIND(myfunctions:custom_concat("First", "last") AS ?concat)
}"""
    response = httpx.get(service_url, params={"query": direct_concat_select}, headers={"accept": "application/json"})
    print(response.text)
    assert response.status_code == 200
    assert response.json()["results"]["bindings"][0]["concat"]["value"] == "Firstlast"


concat_select = """PREFIX myfunctions: <https://w3id.org/sparql-functions/>
SELECT ?concat WHERE {{
    SERVICE <{rdflib_endpoint_url}> {{
        BIND(myfunctions:custom_concat("First", "last") AS ?concat)
    }}
}}"""



import logging
import os
import re
from importlib import resources
from typing import Any, Callable, Dict, List, Optional, Union
from urllib import parse

from fastapi import APIRouter, Query, Request, Response
from fastapi.responses import JSONResponse
from rdflib import RDF, BNode, Dataset, Graph, Literal, URIRef
from rdflib.namespace import DC, RDFS
from rdflib.plugins.sparql import CUSTOM_EVALS, prepareQuery, prepareUpdate
from rdflib.plugins.sparql.evaluate import evalPart
from rdflib.plugins.sparql.evalutils import _eval
from rdflib.plugins.sparql.parserutils import CompValue
from rdflib.plugins.sparql.sparql import QueryContext, SPARQLError
from rdflib.query import Processor

from rdflib_endpoint.utils import (
    API_RESPONSES,
    FORMATS,
    GENERIC_CONTENT_TYPE_TO_RDFLIB_FORMAT,
    GRAPH_CONTENT_TYPE_TO_RDFLIB_FORMAT,
    SD,
    SPARQL_RESULT_CONTENT_TYPE_TO_RDFLIB_FORMAT,
    Defaults,
    QueryExample,
    parse_accept_header,
)


class SparqlRouter(APIRouter):
    """Class to deploy a SPARQL endpoint using a RDFLib Graph."""

    def __init__(
        self,
        *args: Any,
        path: str = "/",
        title: str = Defaults.title,
        description: str = Defaults.description,
        version: str = Defaults.version,
        graph: Union[None, Graph, Dataset] = None,
        service_description: Union[None, Graph] = None,
        processor: Union[str, Processor] = "sparql",
        custom_eval: Optional[Callable[..., Any]] = None,
        functions: Optional[Dict[str, Callable[..., Any]]] = None,
        enable_update: bool = False,
        public_url: str = Defaults.public_url,
        favicon: str = Defaults.favicon,
        example_query: str = Defaults.example,
        example_queries: Optional[Dict[str, QueryExample]] = None,
        **kwargs: Any,
    ) -> None:
        """Create a SPARQL endpoint router.

        The endpoints calls are all defined in this constructor
        """
        self.graph = graph if graph is not None else Dataset(default_union=True)
        """RDFLib Graph for the SPARQL endpoint."""
        self.service_description = service_description if service_description is not None else Graph()
        """RDFLib Graph for the SPARQL Service description."""
        self.functions = functions if functions is not None else {}
        """Custom SPARQL functions to use in the SPARQL queries."""
        self.processor = processor
        """Custom RDFLib SPARQL processor to use for the SPARQL queries."""
        self.title = title
        self.description = description
        self.version = version
        self.path = path
        self.public_url = public_url
        self.example_query = example_query
        self.example_queries = example_queries
        self.example_markdown = f"Example query:\n\n```\n{example_query}\n```"
        self.enable_update = enable_update
        self.favicon = favicon

        # Instantiate APIRouter
        super().__init__(
            *args,
            responses=API_RESPONSES,
            **kwargs,
        )

        # Save custom function in custom evaluation dictionary
        # Handle multiple functions directly in the evalCustomFunctions function
        if custom_eval:
            CUSTOM_EVALS["evalCustomFunctions"] = custom_eval
        elif len(self.functions) > 0:
            CUSTOM_EVALS["evalCustomFunctions"] = self.eval_custom_functions

        self.prepare_sd_graph()

        async def handle_sparql_request(
            request: Request, query: Optional[str] = None, update: Optional[str] = None
        ) -> Response:
            """Handle SPARQL requests to the GET and POST endpoints"""
            if query and update:
                return JSONResponse(
                    status_code=400,
                    content={"message": "Cannot do both query and update"},
                )

            if not query and not update:
                if str(request.headers.get("accept", "")).startswith("text/html"):
                    return self.serve_yasgui()
                # If not asking HTML, return the SPARQL endpoint service description
                if request.headers.get("accept") == "text/turtle":
                    return Response(
                        self.service_description.serialize(format="turtle"),
                        media_type="text/turtle",
                    )
                else:
                    return Response(
                        self.service_description.serialize(format="xml"),
                        media_type="application/xml",
                    )

            # Pretty print the query object
            # from rdflib.plugins.sparql.algebra import pprintAlgebra
            # parsed_query = parser.parseQuery(query)
            # tq = algebraTranslateQuery(parsed_query)
            # pprintAlgebra(tq)

            graph_ns = dict(self.graph.namespaces())

            if query:
                try:
                    parsed_query = prepareQuery(query, initNs=graph_ns)
                    query_results = self.graph.query(parsed_query, processor=self.processor)

                    # Format and return results depending on Accept mime type in request header
                    mime_types = parse_accept_header(request.headers.get("accept", Defaults.content_type))

                    query_operation = re.sub(r"(\w)([A-Z])", r"\1 \2", parsed_query.algebra.name)

                    if query_operation == "Construct Query":
                        content_type_to_rdflib_format = {
                            **GRAPH_CONTENT_TYPE_TO_RDFLIB_FORMAT,
                            **GENERIC_CONTENT_TYPE_TO_RDFLIB_FORMAT,
                        }
                    else:
                        content_type_to_rdflib_format = {
                            **SPARQL_RESULT_CONTENT_TYPE_TO_RDFLIB_FORMAT,
                            **GENERIC_CONTENT_TYPE_TO_RDFLIB_FORMAT,
                        }

                    # Handle cases that are more complicated, like it includes multiple
                    # types, extra information, etc.
                    output_mime_type = Defaults.content_type
                    for mime_type in mime_types:
                        if mime_type in content_type_to_rdflib_format:
                            output_mime_type = mime_type
                            # Use the first mime_type that matches
                            break

                    # Convert generic mime types to specific ones
                    if query_operation == "Construct Query" or query_operation == "Describe Query":
                        if output_mime_type == "text/csv":
                            output_mime_type = "text/turtle"
                        elif output_mime_type == "application/json":
                            output_mime_type = "application/ld+json"
                        elif output_mime_type == "application/xml":
                            output_mime_type = "application/rdf+xml"
                        else:
                            pass

                    if query_operation == "Select Query" or query_operation == "Ask Query":
                        if output_mime_type == "text/csv":
                            output_mime_type = "application/sparql-results+csv"
                        elif output_mime_type == "application/json":
                            output_mime_type = "application/sparql-results+json"
                        elif output_mime_type == "application/xml":
                            output_mime_type = "application/sparql-results+xml"
                        else:
                            pass

                    try:
                        rdflib_format = content_type_to_rdflib_format.get(output_mime_type, output_mime_type)
                        response = Response(
                            query_results.serialize(format=rdflib_format),
                            media_type=output_mime_type,
                        )
                    except Exception as e:
                        logging.error(f"Error serializing the SPARQL query results with RDFLib: {e}")
                        return JSONResponse(
                            status_code=422,
                            content={"message": f"Error serializing the SPARQL query results with RDFLib: {e}"},
                        )
                    else:
                        return response
                except Exception as e:
                    logging.error(f"Error executing the SPARQL query on the RDFLib Graph: {e}")
                    return JSONResponse(
                        status_code=400,
                        content={"message": f"Error executing the SPARQL query on the RDFLib Graph: {e}"},
                    )
            else:  # Update
                if not self.enable_update:
                    return JSONResponse(
                        status_code=403, content={"message": "INSERT and DELETE queries are not allowed."}
                    )
                if rdflib_apikey := os.environ.get("RDFLIB_APIKEY"):
                    authorized = False
                    if auth_header := request.headers.get("Authorization"):  # noqa: SIM102
                        if auth_header.startswith("Bearer ") and auth_header[7:] == rdflib_apikey:
                            authorized = True
                    if not authorized:
                        return JSONResponse(status_code=403, content={"message": "Invalid API KEY."})
                try:
                    prechecked_update: str = update  # type: ignore
                    parsed_update = prepareUpdate(prechecked_update, initNs=graph_ns)
                    self.graph.update(parsed_update, "sparql")
                    return Response(status_code=204)
                except Exception as e:
                    logging.error(f"Error executing the SPARQL update on the RDFLib Graph: {e}")
                    return JSONResponse(
                        status_code=400,
                        content={"message": f"Error executing the SPARQL update on the RDFLib Graph: {e}"},
                    )

        # TODO: use add_api_route? https://github.com/tiangolo/fastapi/blob/d666ccb62216e45ca78643b52c235ba0d2c53986/fastapi/routing.py#L548
        @self.get(
            self.path,
            name="SPARQL endpoint",
            description=self.example_markdown,
            responses=API_RESPONSES,
        )
        async def get_sparql_endpoint(
            request: Request,
            query: Optional[str] = Query(None),
        ) -> Response:
            """Send a SPARQL query to be executed through HTTP GET operation.

            :param request: The HTTP GET request
            :param query: SPARQL query input.
            """
            return await handle_sparql_request(request, query=query)

        @self.post(
            path,
            name="SPARQL endpoint",
            description=self.example_markdown,
            responses=API_RESPONSES,
        )
        async def post_sparql_endpoint(request: Request) -> Response:
            """Send a SPARQL query to be executed through HTTP POST operation.

            :param request: The HTTP POST request with a .body()
            """
            request_body = await request.body()
            body = request_body.decode("utf-8")
            content_type = request.headers.get("content-type", "")
            if "application/sparql-query" in content_type:
                query = body
                update = None
            elif "application/sparql-update" in content_type:
                query = None
                update = body
            elif "application/x-www-form-urlencoded" in content_type:
                request_params = parse.parse_qsl(body)
                query_params = [kvp[1] for kvp in request_params if kvp[0] == "query"]
                query = parse.unquote(query_params[0]) if query_params else None
                update_params = [kvp[1] for kvp in request_params if kvp[0] == "update"]
                update = parse.unquote(update_params[0]) if update_params else None
                # TODO: handle params `using-graph-uri` and `using-named-graph-uri`
                # https://www.w3.org/TR/sparql11-protocol/#update-operation
            elif not body and request.query_params:
                # Blazegraph SERVICE calls uses query_params, not body
                query = parse.unquote(request.query_params.get("query", ""))
                update = request.query_params.get("update")
            else:
                # Response with the service description
                query = None
                update = None
            return await handle_sparql_request(request, query, update)

    def eval_custom_functions(self, ctx: QueryContext, part: CompValue) -> List[Any]:
        """Retrieve variables from a SPARQL-query, then execute registered SPARQL functions
        The results are then stored in Literal objects and added to the query results.

        :param ctx:     <class 'rdflib.plugins.sparql.sparql.QueryContext'>
        :param part:    <class 'rdflib.plugins.sparql.parserutils.CompValue'>
        :return:        <class 'rdflib.plugins.sparql.processor.SPARQLResult'>
        """
        # This part holds basic implementation for adding new functions
        if part.name == "Extend":
            query_results: List[Any] = []
            # Information is retrieved and stored and passed through a generator
            for eval_part in evalPart(ctx, part.p):
                # Checks if the function is a URI (custom function)
                if hasattr(part.expr, "iri"):
                    # Iterate through the custom functions passed in the constructor
                    for function_uri, custom_function in self.functions.items():
                        # Check if URI correspond to a registered custom function
                        if part.expr.iri == URIRef(function_uri):
                            # Execute each function
                            query_results, ctx, part, _ = custom_function(query_results, ctx, part, eval_part)

                else:
                    # For built-in SPARQL functions (that are not URIs)
                    evaluation: List[Any] = [_eval(part.expr, eval_part.forget(ctx, _except=part._vars))]
                    if isinstance(evaluation[0], SPARQLError):
                        raise evaluation[0]
                    # Append results for built-in SPARQL functions
                    for result in evaluation:
                        query_results.append(eval_part.merge({part.var: Literal(result)}))

            return query_results
        raise NotImplementedError()

    def serve_yasgui(self) -> Response:
        """Serve YASGUI interface"""
        import json

        with resources.open_text("rdflib_endpoint", "yasgui.html") as f:
            html_str = f.read()
        html_str = html_str.replace("$TITLE", self.title)
        html_str = html_str.replace("$DESCRIPTION", self.description)
        html_str = html_str.replace("$FAVICON", self.favicon)
        html_str = html_str.replace("$EXAMPLE_QUERY", self.example_query)
        html_str = html_str.replace("$EXAMPLE_QUERIES", json.dumps(self.example_queries))
        return Response(content=html_str, media_type="text/html")

    def prepare_sd_graph(self) -> None:
        """Prepare the endpoint Service Description graph"""
        # https://www.w3.org/TR/sparql12-service-description/
        self.service_description.bind("rdfs", "http://www.w3.org/2000/01/rdf-schema#")
        self.service_description.bind("dc", "http://purl.org/dc/elements/1.1/")
        self.service_description.bind("sd", SD)
        self.service_description.bind("ent", "http://www.w3.org/ns/entailment/")
        self.service_description.bind("prof", "http://www.w3.org/ns/owl-profile/")
        self.service_description.bind("void", "http://rdfs.org/ns/void#")
        self.service_description.bind("formats", FORMATS)
        sd_subj = next(self.service_description.subjects(SD.endpoint, None), None) or BNode()

        if (sd_subj, RDF.type, SD.Service) not in self.service_description:
            self.service_description.add((sd_subj, RDF.type, SD.Service))

        # Check and add endpoint, label and description
        if not any(self.service_description.triples((sd_subj, SD.endpoint, None))):
            self.service_description.add((sd_subj, SD.endpoint, URIRef(self.public_url)))
        if not any(self.service_description.triples((sd_subj, RDFS.label, None))):
            self.service_description.add((sd_subj, RDFS.label, Literal(self.title)))
        if not any(self.service_description.triples((sd_subj, DC.description, None))):
            self.service_description.add((sd_subj, DC.description, Literal(self.description)))

        # Check and add supported language
        if not any(self.service_description.triples((sd_subj, SD.supportedLanguage, SD.SPARQL11Query))):
            self.service_description.add((sd_subj, SD.supportedLanguage, SD.SPARQL11Query))
            if self.enable_update:
                self.service_description.add((sd_subj, SD.supportedLanguage, SD.SPARQL11Update))

        # Add features
        if (
            isinstance(self.graph, Dataset)
            and getattr(self.graph, "default_union", False)
            and not any(self.service_description.triples((sd_subj, SD.feature, SD.BasicFederatedQuery)))
        ):
            self.service_description.add((sd_subj, SD.feature, SD.UnionDefaultGraph))
        if not any(self.service_description.triples((sd_subj, SD.feature, SD.BasicFederatedQuery))):
            self.service_description.add((sd_subj, SD.feature, SD.BasicFederatedQuery))

        # Add result formats
        if not any(self.service_description.triples((sd_subj, SD.resultFormat, FORMATS.SPARQL_Results_JSON))):
            self.service_description.add((sd_subj, SD.resultFormat, FORMATS.SPARQL_Results_JSON))
        if not any(self.service_description.triples((sd_subj, SD.resultFormat, FORMATS.SPARQL_Results_CSV))):
            self.service_description.add((sd_subj, SD.resultFormat, FORMATS.SPARQL_Results_CSV))
        if not any(self.service_description.triples((sd_subj, SD.resultFormat, FORMATS.Turtle))):
            self.service_description.add((sd_subj, SD.resultFormat, FORMATS.Turtle))
        if not any(self.service_description.triples((sd_subj, SD.resultFormat, FORMATS.RDF_XML))):
            self.service_description.add((sd_subj, SD.resultFormat, FORMATS.RDF_XML))

        # Add input formats
        if not any(self.service_description.triples((sd_subj, SD.inputFormat, None))):
            self.service_description.add((sd_subj, SD.inputFormat, FORMATS.RDF_XML))
            self.service_description.add((sd_subj, SD.inputFormat, FORMATS.Turtle))
            self.service_description.add((sd_subj, SD.inputFormat, FORMATS.JSON_LD))
            self.service_description.add((sd_subj, SD.inputFormat, FORMATS.N_Triples))
            self.service_description.add((sd_subj, SD.inputFormat, FORMATS.N_Quads))
            self.service_description.add((sd_subj, SD.inputFormat, FORMATS.TriG))

        # Check and add entailment regime
        # sd_default_entailment_regime = URIRef("http://www.w3.org/ns/sparql-service-description#defaultEntailmentRegime")
        # ent_rdfs = URIRef("http://www.w3.org/ns/entailment/RDFS")
        # if not any(self.service_description.triples((service_uri, sd_default_entailment_regime, ent_rdfs))):
        #     self.service_description.add((service_uri, sd_default_entailment_regime, ent_rdfs))

        # Check if default dataset exists, if not add it
        has_dataset = False
        dataset_node = None
        for _s, _p, o in self.service_description.triples((sd_subj, SD.defaultDataset, None)):
            has_dataset = True
            dataset_node = o
            break

        if not has_dataset:
            dataset_node = BNode()
            self.service_description.add((sd_subj, SD.defaultDataset, dataset_node))
            self.service_description.add((dataset_node, RDF.type, SD.Dataset))

            # Add default graph to the dataset
            graph_node = BNode()
            self.service_description.add((dataset_node, SD.defaultGraph, graph_node))
            self.service_description.add((graph_node, RDF.type, SD.Graph))

            # Add named graphs to the dataset
            if isinstance(self.graph, Dataset):
                # Get the list of distinct graph names with a SPARQL query
                results: Any = self.graph.query("SELECT DISTINCT ?g WHERE { GRAPH ?g { ?s ?p ?o } }")

                for row in results:
                    named_graph_node = BNode()
                    graph_node = BNode()

                    # Add the named graph reference
                    self.service_description.add((dataset_node, SD.namedGraph, named_graph_node))
                    self.service_description.add((named_graph_node, RDF.type, SD.NamedGraph))
                    self.service_description.add((named_graph_node, SD.name, row.g))
                    # self.service_description.add((named_graph_node, SD.entailmentRegime, URIRef("http://www.w3.org/ns/entailment/OWL-RDF-Based")))
                    # self.service_description.add((named_graph_node, SD.supportedEntailmentProfile, URIRef("http://www.w3.org/ns/owl-profile/RL")))

                    # Add graph metadata
                    self.service_description.add((named_graph_node, SD.graph, graph_node))
                    self.service_description.add((graph_node, RDF.type, SD.Graph))

                    # Count triples in the named graph and add it to the description
                    triple_count_query = f"SELECT (COUNT(*) AS ?count) WHERE {{ GRAPH <{row.g}> {{ ?s ?p ?o }} }}"
                    count_result: Any = self.graph.query(triple_count_query)
                    triple_count = next(iter(count_result), [Literal(0)])[0]
                    # result_row: Any = next(iter(count_result), None)
                    # triple_count = result_row.count if result_row else Literal(0)
                    self.service_description.add((graph_node, URIRef("http://rdfs.org/ns/void#triples"), triple_count))

        # Add custom functions to the service description
        for custom_function_uri in self.functions:
            function_uri = URIRef(custom_function_uri)

            if (function_uri, RDF.type, SD.Function) not in self.service_description:
                self.service_description.add((function_uri, RDF.type, SD.Function))
            if (sd_subj, SD.extensionFunction, function_uri) not in self.service_description:
                self.service_description.add((sd_subj, SD.extensionFunction, function_uri))
        # return self.service_description

        # Add custom functions URI to the service description
        for custom_function_uri in self.functions:
            self.service_description.add(
                (
                    URIRef(custom_function_uri),
                    RDF.type,
                    SD.Function,
                )
            )
            self.service_description.add(
                (
                    sd_subj,
                    SD.extensionFunction,
                    URIRef(custom_function_uri),
                )
            )



import pytest

import rdflib_endpoint.sparql_router

accept_cases = [
    ("text/xml", "text/xml"),
    ("text/rdf+xml, text/xml, */*", "text/rdf+xml"),
    ("text/html, application/xhtml+xml, application/xml;q=0.9, */*;q=0.8", "text/html"),
    ("text/html, application/xhtml+xml, application/xml;q=0.9, */*;q=", "text/html"),
    ("text/html;q=0.3, application/xhtml+xml, application/xml;q=0.9, */*;q=0.8", "application/xhtml+xml"),
    (
        'text/turtle;q=0.9;profile="urn:example:profile-1", text/turtle;q=0.7;profile="urn:example:profile-2"',
        "text/turtle",
    ),
]


@pytest.mark.parametrize("accept,expected", accept_cases)
def test_accept_preference(accept, expected):
    pref = rdflib_endpoint.sparql_router.parse_accept_header(accept)
    assert pref[0] == expected



from typing import Any, Dict, List, Optional, TypedDict, Union

from rdflib import Namespace

SD = Namespace("http://www.w3.org/ns/sparql-service-description#")
FORMATS = Namespace("http://www.w3.org/ns/formats/")


class Defaults:
    """Default configuration values for the SPARQL endpoint."""

    title: str = "SPARQL endpoint for RDFLib graph"
    description: str = "A SPARQL endpoint to serve machine learning models, or any other logic implemented in Python. \n[Source code](https://github.com/vemonet/rdflib-endpoint)"
    version: str = "0.1.0"
    public_url: str = "https://your-endpoint/sparql"
    favicon: str = "https://rdflib.readthedocs.io/en/stable/_static/RDFlib.png"
    example: str = """\
PREFIX myfunctions: <https://w3id.org/sparql-functions/>

SELECT ?concat ?concatLength WHERE {
    BIND("First" AS ?first)
    BIND(myfunctions:custom_concat(?first, "last") AS ?concat)
}
""".rstrip()
    #: This is default for federated queries
    content_type: str = "application/xml"


API_RESPONSES: Optional[Dict[Union[int, str], Dict[str, Any]]] = {
    200: {
        "description": "SPARQL query results",
        "content": {
            "application/sparql-results+json": {"example": {"results": {"bindings": []}, "head": {"vars": []}}},
            "application/json": {"example": {"results": {"bindings": []}, "head": {"vars": []}}},
            "text/csv": {"example": "s,p,o"},
            "application/sparql-results+csv": {"example": "s,p,o"},
            "application/sparql-results+xml": {"example": "<root></root>"},
            "application/xml": {"example": "<root></root>"},
            "text/turtle": {"example": "<http://subject> <http://predicate> <http://object> ."},
            "application/n-triples": {"example": "<http://subject> <http://predicate> <http://object> ."},
            "text/n3": {"example": "<http://subject> <http://predicate> <http://object> ."},
            "application/n-quads": {"example": "<http://subject> <http://predicate> <http://object> <http://graph> ."},
            "application/trig": {
                "example": "GRAPH <http://graph> {<http://subject> <http://predicate> <http://object> .}"
            },
            "application/trix": {"example": "<xml></xml>"},
            "application/ld+json": {
                "example": [
                    {
                        "@id": "http://subject",
                        "@type": ["http://object"],
                        "http://www.w3.org/2000/01/rdf-schema#label": [{"@value": "foo"}],
                    }
                ]
            },
            # "application/rdf+xml": {
            #     "example": '<?xml version="1.0" encoding="UTF-8"?> <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"></rdf:RDF>'
            # },
        },
    },
    400: {
        "description": "Bad Request",
    },
    403: {
        "description": "Forbidden",
    },
    422: {
        "description": "Unprocessable Entity",
    },
}

#: A mapping from content types to the keys used for serializing
#: in :meth:`rdflib.Graph.serialize` and other serialization functions
GRAPH_CONTENT_TYPE_TO_RDFLIB_FORMAT = {
    # https://www.w3.org/TR/rdf-sparql-XMLres/
    "application/rdf+xml": "xml",  # for compatibility
    "application/ld+json": "json-ld",
    # https://www.w3.org/TR/sparql11-results-csv-tsv/
    # Extras
    "text/turtle": "ttl",
    "text/n3": "n3",
    "application/n-triples": "nt",
    "text/plain": "nt",
    "application/trig": "trig",
    "application/trix": "trix",
    "application/n-quads": "nquads",
}

SPARQL_RESULT_CONTENT_TYPE_TO_RDFLIB_FORMAT = {
    # https://www.w3.org/TR/sparql11-results-json/
    "application/sparql-results+json": "json",
    # https://www.w3.org/TR/rdf-sparql-XMLres/
    "application/sparql-results+xml": "xml",
    # https://www.w3.org/TR/sparql11-results-csv-tsv/
    "application/sparql-results+csv": "csv",
}

GENERIC_CONTENT_TYPE_TO_RDFLIB_FORMAT = {
    # https://www.w3.org/TR/sparql11-results-json/
    "application/json": "json",
    "text/json": "json",
    # https://www.w3.org/TR/rdf-sparql-XMLres/
    "application/xml": "xml",  # for compatibility
    "text/xml": "xml",  # not standard
    # https://www.w3.org/TR/sparql11-results-csv-tsv/
    "text/csv": "csv",  # for compatibility
}


def parse_accept_header(accept: str) -> List[str]:
    """Given an accept header string, return a list of media types in order of preference.

    :param accept: Accept header value
    :return: Ordered list of media type preferences
    """

    def _parse_preference(qpref: str) -> float:
        qparts = qpref.split("=")
        try:
            return float(qparts[1].strip())
        except (ValueError, IndexError):
            pass
        return 1.0

    preferences = []
    types = accept.split(",")
    dpref = 2.0
    for mtype in types:
        parts = mtype.split(";")
        parts = [part.strip() for part in parts]
        pref = dpref
        try:
            for part in parts[1:]:
                if part.startswith("q="):
                    pref = _parse_preference(part)
                    break
        except IndexError:
            pass
        # preserve order of appearance in the list
        dpref = dpref - 0.01
        preferences.append((parts[0], pref))
    preferences.sort(key=lambda x: -x[1])
    return [pref[0] for pref in preferences]


class QueryExample(TypedDict, total=False):
    """Dictionary to store example queries for the SPARQL endpoint."""

    query: str
    endpoint: Optional[str]



"""A package to deploy SPARQL endpoint to serve local RDF files, machine learning models, or any other logic implemented in Python, using RDFLib and FastAPI."""

__version__ = "0.5.4"

from .sparql_router import SparqlRouter
from .sparql_endpoint import SparqlEndpoint

__all__ = [
    "SparqlEndpoint",
    "SparqlRouter",
]



from example.main import app
from fastapi.testclient import TestClient
from rdflib import Graph

from rdflib_endpoint.sparql_router import SD

# Use app defined in example folder
endpoint = TestClient(app)


def test_service_description():
    # Check GET turtle
    response = endpoint.get("/", headers={"accept": "text/turtle"})
    print(response.text)
    assert response.status_code == 200
    g = Graph()
    g.parse(data=response.text, format="turtle")
    assert any(g.triples((None, SD.endpoint, None))), "Missing sd:endpoint in service description"
    assert any(g.triples((None, SD.extensionFunction, None))), "Missing sd:extensionFunction in service description"
    assert len(list(g.triples((None, SD.extensionFunction, None)))) == 2, "Expected exactly 2 extension functions"
    assert len(list(g.triples((None, SD.namedGraph, None)))) == 2, "Expected exactly 2 named graphs"

    # Check POST turtle
    response = endpoint.post("/", headers={"accept": "text/turtle"})
    assert response.status_code == 200
    g = Graph()
    g.parse(data=response.text, format="turtle")
    assert any(g.triples((None, SD.endpoint, None))), "Missing sd:endpoint in service description"
    assert any(g.triples((None, SD.extensionFunction, None))), "Missing sd:extensionFunction in service description"
    assert len(list(g.triples((None, SD.extensionFunction, None)))) == 2, "Expected exactly 2 extension functions"

    # Check POST XML
    response = endpoint.post("/", headers={"accept": "application/xml"})
    assert response.status_code == 200
    g = Graph()
    g.parse(data=response.text, format="xml")
    assert any(g.triples((None, SD.endpoint, None))), "Missing sd:endpoint in service description"
    assert any(g.triples((None, SD.extensionFunction, None))), "Missing sd:extensionFunction in service description"
    assert len(list(g.triples((None, SD.extensionFunction, None)))) == 2, "Expected exactly 2 extension functions"


def test_custom_concat():
    response = endpoint.get("/", params={"query": custom_concat_query}, headers={"accept": "application/json"})
    # print(response.json())
    assert response.status_code == 200
    assert response.json()["results"]["bindings"][0]["concat"]["value"] == "Firstlast"

    response = endpoint.post(
        "/",
        data={"query": custom_concat_query},
        headers={"accept": "application/json"},
    )
    assert response.status_code == 200
    assert response.json()["results"]["bindings"][0]["concat"]["value"] == "Firstlast"


def test_bad_request():
    response = endpoint.get("/?query=figarofigarofigaro", headers={"accept": "application/json"})
    assert response.status_code == 400


custom_concat_query = """PREFIX myfunctions: <https://w3id.org/sparql-functions/>
SELECT ?concat ?concatLength WHERE {
    BIND("First" AS ?first)
    BIND(myfunctions:custom_concat(?first, "last") AS ?concat)
}"""



import rdflib
from fastapi.testclient import TestClient
from rdflib import URIRef
from rdflib.namespace import RDF, RDFS
from rdflib.plugins.sparql.evaluate import evalBGP

from rdflib_endpoint import SparqlEndpoint

# TODO: not used due to bug with FastAPI TestClient when using different apps in the tests


def custom_eval(ctx, part):
    """Rewrite triple patterns to get super-classes"""
    if part.name == "BGP":
        # rewrite triples
        triples = []
        for t in part.triples:
            if t[1] == RDF.type:
                bnode = rdflib.BNode()
                triples.append((t[0], t[1], bnode))
                triples.append((bnode, RDFS.subClassOf, t[2]))
            else:
                triples.append(t)
        # delegate to normal evalBGP
        return evalBGP(ctx, triples)
    raise NotImplementedError()


g = rdflib.Graph()
g.add((URIRef("http://human"), RDFS.subClassOf, URIRef("http://mammal")))
g.add((URIRef("http://alice"), RDF.type, URIRef("http://human")))

eval_app = SparqlEndpoint(graph=g, custom_eval=custom_eval, functions={})
eval_endpoint = TestClient(eval_app)


def test_custom_eval():
    # eval_app = SparqlEndpoint(
    #     graph=g,
    #     custom_eval=custom_eval,
    #     functions={}
    # )
    # eval_endpoint = TestClient(eval_app)

    response = eval_endpoint.get("/?query=" + select_parent, headers={"accept": "application/json"})
    print(response.json())
    assert response.status_code == 200
    print(response.json()["results"]["bindings"])
    assert str(response.json()["results"]["bindings"][0]["s"]["value"]) == "http://alice"

    response = eval_endpoint.post("/", data="query=" + select_parent, headers={"accept": "application/json"})
    assert response.status_code == 200
    assert str(response.json()["results"]["bindings"][0]["s"]["value"]) == "http://alice"


select_parent = """SELECT * WHERE {
    ?s a <http://mammal> .
}"""



from fastapi.testclient import TestClient
from rdflib import RDF, RDFS, Graph, Literal, URIRef

from rdflib_endpoint import SparqlEndpoint

g = Graph(store="Oxigraph")
g.add((URIRef("http://subject"), RDF.type, URIRef("http://object")))
g.add((URIRef("http://subject"), RDFS.label, Literal("test value")))


app = SparqlEndpoint(graph=g)

endpoint = TestClient(app)


def test_service_description():
    response = endpoint.get("/", headers={"accept": "text/turtle"})
    # print(response.text.strip())
    assert response.status_code == 200

    response = endpoint.post("/", headers={"accept": "text/turtle"})
    assert response.status_code == 200

    # Check for application/xml
    response = endpoint.post("/", headers={"accept": "application/xml"})
    assert response.status_code == 200


def test_custom_concat_json():
    response = endpoint.get("/?query=", params={"query": label_select}, headers={"accept": "application/json"})
    assert response.status_code == 200
    assert response.json()["results"]["bindings"][0]["label"]["value"] == "test value"

    response = endpoint.post("/", data={"query": label_select}, headers={"accept": "application/json"})
    assert response.status_code == 200
    assert response.json()["results"]["bindings"][0]["label"]["value"] == "test value"


def test_select_noaccept_xml():
    response = endpoint.post("/", data={"query": label_select})
    assert response.status_code == 200
    # assert response.json()['results']['bindings'][0]['concat']['value'] == "Firstlast"


def test_select_csv():
    response = endpoint.post("/", data={"query": label_select}, headers={"accept": "text/csv"})
    assert response.status_code == 200
    # assert response.json()['results']['bindings'][0]['concat']['value'] == "Firstlast"


def test_select_turtle():
    response = endpoint.post("/", data={"query": label_select}, headers={"accept": "text/turtle"})
    assert response.status_code == 200
    assert response.headers["content-type"] == "application/sparql-results+xml"
    # assert response.json()['results']['bindings'][0]['concat']['value'] == "Firstlast"


def test_concat_construct_turtle():
    # expected to return turtle
    response = endpoint.post(
        "/",
        data="query=" + label_construct,
        headers={"accept": "application/json"},
    )
    assert response.status_code == 200
    # assert response.json()['results']['bindings'][0]['concat']['value'] == "Firstlast"


def test_concat_construct_xml():
    # expected to return turtle
    response = endpoint.post(
        "/",
        data="query=" + label_construct,
        headers={"accept": "application/xml"},
    )
    assert response.status_code == 200


def test_bad_request():
    response = endpoint.get("/?query=figarofigarofigaro", headers={"accept": "application/json"})
    assert response.status_code == 400


label_select = """PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
SELECT ?label WHERE {
    ?s rdfs:label ?label .
}"""


label_construct = """PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
CONSTRUCT {
    <http://test> <http://label> ?label .
} WHERE {
    ?s rdfs:label ?label .
}"""



import glob
import os
import tempfile
from unittest.mock import MagicMock, patch

from click.testing import CliRunner

from rdflib_endpoint.__main__ import cli

runner = CliRunner()

out_formats = ["ttl", "nt", "xml", "jsonld", "trig"]


def test_convert():
    for out_format in out_formats:
        with tempfile.NamedTemporaryFile(delete=True) as tmp_file:
            out_file = str(f"{tmp_file}.{out_format}")
            result = runner.invoke(
                cli,
                ["convert", "tests/resources/test2.ttl", "--output", out_file],
            )
            assert result.exit_code == 0
            with open(out_file) as file:
                content = file.read()
                assert len(content) > 1

    # Fix issue with python creating unnecessary temp files on disk
    for f in glob.glob("<tempfile._TemporaryFileWrapper*"):
        os.remove(f)


def test_convert_oxigraph():
    with tempfile.NamedTemporaryFile(delete=True) as tmp_file:
        result = runner.invoke(
            cli,
            [
                "convert",
                "--store",
                "oxigraph",
                "tests/resources/test2.ttl",
                "--output",
                str(tmp_file),
            ],
        )
        assert result.exit_code == 0
        with open(str(tmp_file)) as file:
            content = file.read()
            assert len(content) > 1
    # Fix issue with python creating unnecessary temp files on disk
    for f in glob.glob("<tempfile._TemporaryFileWrapper*"):
        os.remove(f)


# NOTE: Needs to run last tests, for some reason patching uvicorn as a side effects on follow up tests


@patch("rdflib_endpoint.__main__.uvicorn.run")
def test_serve(mock_run: MagicMock) -> None:
    """Test serve, mock uvicorn.run to prevent API hanging"""
    mock_run.return_value = None
    result = runner.invoke(
        cli,
        [
            "serve",
            "tests/resources/test.nq",
            "tests/resources/test2.ttl",
            "tests/resources/another.jsonld",
        ],
    )
    assert result.exit_code == 0


@patch("rdflib_endpoint.__main__.uvicorn.run")
def test_serve_oxigraph(mock_run: MagicMock) -> None:
    """Test serve oxigraph, mock uvicorn.run to prevent API hanging"""
    mock_run.return_value = None
    result = runner.invoke(
        cli,
        [
            "serve",
            "--store",
            "oxigraph",
            "tests/resources/test.nq",
            "tests/resources/test2.ttl",
            "tests/resources/another.jsonld",
        ],
    )
    assert result.exit_code == 0



import time
from typing import Any, Callable, Dict, Optional, Union

from fastapi import FastAPI, Request, Response
from fastapi.middleware.cors import CORSMiddleware
from rdflib import Dataset, Graph
from rdflib.query import Processor

from rdflib_endpoint.sparql_router import SparqlRouter
from rdflib_endpoint.utils import Defaults, QueryExample


class SparqlEndpoint(FastAPI):
    """Class to deploy a SPARQL endpoint using a RDFLib Graph."""

    def __init__(
        self,
        *args: Any,
        path: str = "/",
        title: str = Defaults.title,
        description: str = Defaults.description,
        version: str = Defaults.version,
        graph: Union[None, Graph, Dataset] = None,
        functions: Optional[Dict[str, Callable[..., Any]]] = None,
        processor: Union[str, Processor] = "sparql",
        custom_eval: Optional[Callable[..., Any]] = None,
        enable_update: bool = False,
        cors_enabled: bool = True,
        public_url: str = Defaults.public_url,
        example_query: str = Defaults.example,
        example_queries: Optional[Dict[str, QueryExample]] = None,
        favicon: str = Defaults.favicon,
        **kwargs: Any,
    ) -> None:
        """
        Create a SPARQL endpoint.

        It's mainly a wrapper to include the SPARQL router. Everything happens in the router.
        """
        self.title = title
        self.description = description
        self.version = version

        # Instantiate FastAPI
        super().__init__(
            *args,
            title=title,
            description=description,
            version=version,
            **kwargs,
        )

        sparql_router = SparqlRouter(
            path=path,
            title=title,
            description=description,
            version=version,
            graph=graph,
            functions=functions,
            processor=processor,
            custom_eval=custom_eval,
            enable_update=enable_update,
            public_url=public_url,
            example_query=example_query,
            example_queries=example_queries,
            favicon=favicon,
        )
        self.include_router(sparql_router)

        if cors_enabled:
            self.add_middleware(
                CORSMiddleware,
                allow_origins=["*"],
                allow_credentials=True,
                allow_methods=["*"],
                allow_headers=["*"],
            )

        @self.middleware("http")
        async def add_process_time_header(request: Request, call_next: Any) -> Response:
            """Add a Server-Timing header with the total processing time for each request."""
            start_time = time.time()
            response: Response = await call_next(request)
            response.headers["Server-Timing"] = f"total;dur={time.time() - start_time}"
            return response



import pytest
from example.main import custom_concat
from fastapi.testclient import TestClient
from rdflib import RDFS, Graph, Literal, URIRef

from rdflib_endpoint import SparqlEndpoint
from rdflib_endpoint.sparql_router import SD

# graph = Dataset(default_union=False)
graph = Graph()


@pytest.fixture(autouse=True)
def clear_graph():
    # Workaround to clear graph without putting
    # graph, app and endpoint into a fixture
    # and modifying the test fixture usage.
    for triple in graph:
        graph.remove(triple)


app = SparqlEndpoint(
    graph=graph,
    functions={
        "https://w3id.org/sparql-functions/custom_concat": custom_concat,
    },
    enable_update=True,
)

endpoint = TestClient(app)


def test_service_description():
    # Check GET turtle
    response = endpoint.get("/", headers={"accept": "text/turtle"})
    assert response.status_code == 200
    g = Graph()
    g.parse(data=response.text, format="turtle")
    assert any(g.triples((None, SD.endpoint, None))), "Missing sd:endpoint in service description"
    assert any(g.triples((None, SD.extensionFunction, None))), "Missing sd:extensionFunction in service description"
    assert len(list(g.triples((None, SD.extensionFunction, None)))) == 1, "Expected exactly 1 extension function"

    # Check POST XML
    response = endpoint.post("/", headers={"accept": "application/xml"})
    assert response.status_code == 200
    g = Graph()
    g.parse(data=response.text, format="xml")
    assert any(g.triples((None, SD.endpoint, None))), "Missing sd:endpoint in service description"
    assert any(g.triples((None, SD.extensionFunction, None))), "Missing sd:extensionFunction in service description"
    assert len(list(g.triples((None, SD.extensionFunction, None)))) == 1, "Expected exactly 1 extension function"


def test_custom_concat_json():
    response = endpoint.get("/", params={"query": concat_select}, headers={"accept": "application/json"})
    # print(response.json())
    assert response.status_code == 200
    assert response.json()["results"]["bindings"][0]["concat"]["value"] == "Firstlast"

    response = endpoint.post("/", data={"query": concat_select}, headers={"accept": "application/json"})
    assert response.status_code == 200
    assert response.json()["results"]["bindings"][0]["concat"]["value"] == "Firstlast"

    response = endpoint.post(
        "/", data=concat_select, headers={"accept": "application/json", "content-type": "application/sparql-query"}
    )
    assert response.status_code == 200
    assert response.json()["results"]["bindings"][0]["concat"]["value"] == "Firstlast"


def test_select_noaccept_xml():
    response = endpoint.post("/", data={"query": concat_select})
    assert response.status_code == 200
    assert response.text.startswith("<?xml ")


def test_select_csv():
    response = endpoint.post("/", data={"query": concat_select}, headers={"accept": "text/csv"})
    assert response.status_code == 200


label_patch = """
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
DELETE { ?subject rdfs:label "foo" }
INSERT { ?subject rdfs:label "bar" }
WHERE { ?subject rdfs:label "foo" }
"""


@pytest.mark.parametrize(
    "api_key,key_provided,param_method",
    [
        (api_key, key_provided, param_method)
        for api_key in [None, "key"]
        for key_provided in [True, False]
        for param_method in ["body_form", "body_direct"]
    ],
)
def test_sparql_update(api_key, key_provided, param_method, monkeypatch):
    if api_key:
        monkeypatch.setenv("RDFLIB_APIKEY", api_key)
    subject = URIRef("http://server.test/subject")
    headers = {}
    if key_provided:
        headers["Authorization"] = "Bearer key"
    graph.add((subject, RDFS.label, Literal("foo")))
    if param_method == "body_form":
        request_args = {"data": {"update": label_patch}}
    else:
        # direct
        headers["content-type"] = "application/sparql-update"
        request_args = {"data": label_patch}
    response = endpoint.post("/", headers=headers, **request_args)
    if api_key is None or key_provided:
        assert response.status_code == 204
        assert (subject, RDFS.label, Literal("foo")) not in graph
        assert (subject, RDFS.label, Literal("bar")) in graph
    else:
        assert response.status_code == 403
        assert (subject, RDFS.label, Literal("foo")) in graph
        assert (subject, RDFS.label, Literal("bar")) not in graph


def test_sparql_query_update_fail():
    response = endpoint.post("/", data={"update": label_patch, "query": label_patch})
    assert response.status_code == 400


def test_multiple_accept_return_json():
    response = endpoint.get(
        "/",
        params={"query": concat_select},
        headers={"accept": "text/html;q=0.3, application/xml;q=0.9, application/json, */*;q=0.8"},
    )
    assert response.status_code == 200
    assert response.json()["results"]["bindings"][0]["concat"]["value"] == "Firstlast"


def test_multiple_accept_return_json2():
    response = endpoint.get(
        "/",
        params={"query": concat_select},
        headers={"accept": "text/html;q=0.3, application/json, application/xml;q=0.9, */*;q=0.8"},
    )
    assert response.status_code == 200
    assert response.json()["results"]["bindings"][0]["concat"]["value"] == "Firstlast"


def test_select_turtle():
    response = endpoint.post("/", data={"query": concat_select}, headers={"accept": "text/turtle"})
    assert response.status_code == 200
    assert response.headers["content-type"] == "application/sparql-results+xml"


def test_concat_construct_turtle():
    response = endpoint.post(
        "/",
        data={"query": custom_concat_construct},
        headers={"accept": "text/turtle"},
    )
    assert response.status_code == 200
    assert response.text.startswith("@prefix ")


def test_concat_construct_csv():
    response = endpoint.post(
        "/",
        data={"query": custom_concat_construct},
        headers={"accept": "text/csv"},
    )
    assert response.status_code == 200
    assert response.text.startswith("@prefix ")


def test_concat_construct_jsonld():
    response = endpoint.post(
        "/",
        data={"query": custom_concat_construct},
        headers={"accept": "application/json"},
    )
    assert response.status_code == 200
    assert response.json()[0]["@id"] == "http://example.com/test"


def test_concat_construct_xml():
    # expected to return turtle
    response = endpoint.post(
        "/",
        data={"query": custom_concat_construct},
        headers={"accept": "application/xml"},
    )
    assert response.status_code == 200
    assert response.text.startswith("<?xml ")


def test_concat_construct_rdf_xml_with_two_accept_options():
    response = endpoint.post(
        "/",
        data={"query": custom_concat_construct},
        headers={"accept": "application/sparql-results+xml, application/rdf+xml"},
    )
    assert response.status_code == 200
    assert response.headers["content-type"] == "application/rdf+xml"
    assert response.text.startswith("<?xml ")
    assert "<rdf:RDF" in response.text


def test_yasgui():
    # expected to return turtle
    response = endpoint.get(
        "/",
        headers={"accept": "text/html"},
    )
    assert response.status_code == 200


def test_bad_request():
    response = endpoint.get("/?query=figarofigarofigaro", headers={"accept": "application/json"})
    assert response.status_code == 400


concat_select = """PREFIX myfunctions: <https://w3id.org/sparql-functions/>
SELECT ?concat ?concatLength WHERE {
    BIND("First" AS ?first)
    BIND(myfunctions:custom_concat(?first, "last") AS ?concat)
}"""

custom_concat_construct = """PREFIX myfunctions: <https://w3id.org/sparql-functions/>
CONSTRUCT {
    <http://example.com/test> <http://example.com/concat> ?concat, ?concatLength .
} WHERE {
    BIND("First" AS ?first)
    BIND(myfunctions:custom_concat(?first, "last") AS ?concat)
}"""
